/************************************************************************************************

  Asteroids.js
  Original Java Code Copyright 1998 by Mike Hall.  Used by permission.

  Keyboard Controls:

  S            - Start Game    P           - Pause Game
  Cursor Left  - Rotate Left   Cursor Up   - Fire Thrusters
  Cursor Right - Rotate Right  Cursor Down - Fire Retro Thrusters
  Spacebar     - Fire Cannon   H           - Hyperspace
  M            - Toggle Sound  D           - Toggle Graphics Detail

************************************************************************************************/

/* Polygon class */

function Polygon() {
    this.xpoints = [];
    this.ypoints = [];
    this.npoints = 0;
    this.polygon = new QPolygon();
}

Polygon.prototype.addPoint = function(x, y) {
    this.xpoints.push(x);
    this.ypoints.push(y);
    this.polygon.append(new QPoint(x, y));
    this.npoints = this.xpoints.length;
}

Polygon.prototype.inside = function(x, y) {

  var inside=false;
  var j = 0;

  for (var i=0; i< this.npoints; i++) {
    j++;
    if (j == this.npoints) j=0;
    if (this.ypoints[i]<y && this.ypoints[j]>=y
     || this.ypoints[j]<y && this.ypoints[i]>=y) {
      if (this.xpoints[i]+(y-this.ypoints[i])/(this.ypoints[j]-this.ypoints[i])*(this.xpoints[j]-this.xpoints[i])<x) {
        inside=!inside;
      }
    }
  }

  return inside;
}

Polygon.prototype.draw = function(painter) {
    painter.drawPolygon(this.polygon);
}

Polygon.prototype.fill = function(painter) {
    // Not needed in Qt, because our 'setBrush' setting
    // automatically makes the draw operation fill the polygon
    // painter.drawPolygon(this.polygon, Qt.WindingFill);
}

/* Point class */
function Point(xVal, yVal) {
    this.x = xVal;
    this.y = yVal;
}

/************************************************************************************************
  The AsteroidsSprite class defines a game object, including it's shape, position, movement and
  rotation. It also can detemine if two objects collide.
************************************************************************************************/

function AsteroidsSprite(owner) {

  // Fields:
  /* Polygon */ this.shape = new Polygon();  // Initial sprite shape, centered at the origin (0,0).
  /* boolean */ this.active = false;         // Active flag.
  /* double */  this.angle = 0;              // Current angle of rotation.
  /* double */  this.deltaAngle = 0;         // Amount to change the rotation angle.
  /* double */  this.currentX = 0;           // Current position on screen.
  /* double */  this.currentY = 0;
  /* double */  this.deltaX = 0;             // Amount to change the screen position.
  /* double */  this.deltaY = 0;
  /* Polygon */ this.sprite = new Polygon(); // Final location and shape of sprite after applying rotation and
                                             // moving to screen position. Used for drawing on the screen and
                                             // in detecting collisions.
    this.owner = owner;                      // The Asteroids widget that owns
                                             // this sprite.
}

  // Methods:

  AsteroidsSprite.prototype.advance = function() {

    // Update the rotation and position of the sprite based on the delta values. If the sprite
    // moves off the edge of the screen, it is wrapped around to the other side.

    this.angle += this.deltaAngle;
    if (this.angle < 0)
      this.angle += 2 * Math.PI;
    if (this.angle > 2 * Math.PI)
      this.angle -= 2 * Math.PI;
    this.currentX += this.deltaX;
    if (this.currentX < -this.owner.screen_width / 2)
      this.currentX += this.owner.screen_width;
    if (this.currentX > this.owner.screen_width / 2)
      this.currentX -= this.owner.screen_width;
    this.currentY -= this.deltaY;
    if (this.currentY < -this.owner.screen_height / 2)
      this.currentY += this.owner.screen_height;
    if (this.currentY > this.owner.screen_height / 2)
      this.currentY -= this.owner.screen_height;
  }

  AsteroidsSprite.prototype.render = function() {

    // Render the sprite's shape and location by rotating it's base shape and moving it to
    // it's proper screen position.

    this.sprite = new Polygon();
    for (var i = 0; i < this.shape.npoints; i++)
      this.sprite.addPoint(Math.round(this.shape.xpoints[i] * Math.cos(this.angle) + this.shape.ypoints[i] * Math.sin(this.angle)) + Math.round(this.currentX) + this.owner.screen_width / 2,
                 Math.round(this.shape.ypoints[i] * Math.cos(this.angle) - this.shape.xpoints[i] * Math.sin(this.angle)) + Math.round(this.currentY) + this.owner.screen_height / 2);
  }

  AsteroidsSprite.prototype.isColliding = function(/* AsteroidsSprite */ s) {

    var i;

    // Determine if one sprite overlaps with another, i.e., if any vertice
    // of one sprite lands inside the other.

    for (i = 0; i < s.sprite.npoints; i++)
      if (this.sprite.inside(s.sprite.xpoints[i], s.sprite.ypoints[i]))
        return true;
    for (i = 0; i < this.sprite.npoints; i++)
      if (s.sprite.inside(this.sprite.xpoints[i], this.sprite.ypoints[i]))
        return true;
    return false;
  }


Asteroids.prototype = new QWidget();

/************************************************************************************************
  Main application code -- constants and variables
************************************************************************************************/

// Constants
Asteroids.prototype.DELAY = 50;             // Milliseconds between screen updates.

Asteroids.prototype.MAX_SHIPS = 3;           // Starting number of ships per game.
Asteroids.prototype.MAX_SHOTS =  6;          // Maximum number of sprites for photons,
Asteroids.prototype.MAX_ROCKS =  8;          // asteroids and explosions.
Asteroids.prototype.MAX_SCRAP = 20;

Asteroids.prototype.SCRAP_COUNT = 30;        // Counter starting values.
Asteroids.prototype.HYPER_COUNT = 60;
Asteroids.prototype.STORM_PAUSE = 30;
Asteroids.prototype.UFO_PASSES  =  3;

Asteroids.prototype.MIN_ROCK_SIDES =  8;     // Asteroid shape and size ranges.
Asteroids.prototype.MAX_ROCK_SIDES = 12;
Asteroids.prototype.MIN_ROCK_SIZE  = 20;
Asteroids.prototype.MAX_ROCK_SIZE  = 40;
Asteroids.prototype.MIN_ROCK_SPEED =  2;
Asteroids.prototype.MAX_ROCK_SPEED = 12;

Asteroids.prototype.BIG_POINTS     =  25;     // Points for shooting different objects.
Asteroids.prototype.SMALL_POINTS   =  50;
Asteroids.prototype.UFO_POINTS     = 250;
Asteroids.prototype.MISSILE_POINTS = 500;

Asteroids.prototype.NEW_SHIP_POINTS = 5000;  // Number of points needed to earn a new ship.
Asteroids.prototype.NEW_UFO_POINTS  = 2750;  // Number of points between flying saucers.

Asteroids.prototype.BLACK_COLOR = new QColor(0, 0, 0);
Asteroids.prototype.YELLOW_COLOR = new QColor(255, 255, 0);


  // Font data.
  /* NOTE: Fonts are not supported yet in the JavaScript version
  Font font = new Font("Helvetica", Font.BOLD, 12);
  FontMetrics fm;
  */
Asteroids.prototype.FONT_WIDTH = 5; // Just an approximation of average width
Asteroids.prototype.FONT_HEIGHT = 18; // getStringHeight("X");


/************************************************************************************************
  Main application code -- Constructor.
************************************************************************************************/


function Asteroids(parent)
{

    QWidget.call(this, parent);

    // Dimensions of the graphics area.
    this.screen_width = 600;
    this.screen_height = 400;

    // Background stars.
    this.numStars = 0;
    this.stars = [];

    // Game data.
    this.score = 0;
    this.highScore = 0;
    this.newShipScore = 0;
    this.newUfoScore = 0;

    this.loaded = false;
    this.paused = false;
    this.playing = false;
    this.sound = false;
    this.detail = false;

    // Key flags.
    this.left  = false;
    this.right = false;
    this.up    = false;
    this.down  = false;

    // Sprite objects.
    this.ship = null;
    this.ufo = null;
    this.missile = null;
    this.photons    = []; /* new AsteroidsSprite[MAX_SHOTS]; */
    this.asteroids  = []; /* new AsteroidsSprite[MAX_ROCKS]; */
    this.explosions = []; /* new AsteroidsSprite[MAX_SCRAP]; */

    // Ship data.
    this.shipsLeft = 0;       // Number of ships left to play, including current one.
    this.shipCounter = 0;     // Time counter for ship explosion.
    this.hyperCounter = 0;    // Time counter for hyperspace.

    // Photon data.
    this.photonCounter = []; /* new int[MAX_SHOTS]; */ // Time counter for life of a photon.
    this.photonIndex = 0;                              // Next available photon sprite.

    // Flying saucer data.
    this.ufoPassesLeft = 0;    // Number of flying saucer passes.
    this.ufoCounter = 0;       // Time counter for each pass.

    // Missile data.
    this.missileCounter;    // Counter for life of missile.

    // Asteroid data.
    this.asteroidIsSmall = [] /* new boolean[MAX_ROCKS]; */ // Asteroid size flag.
    this.asteroidsCounter = 0;                              // Break-time counter.
    this.asteroidsSpeed = 0;                                // Asteroid speed.
    this.asteroidsLeft = 0;                                 // Number of active asteroids.

    // Explosion data.
    this.explosionCounter = []; /* new int[MAX_SCRAP]; */ // Time counters for explosions.
    this.explosionIndex = 0;                              // Next available explosion sprite.

    // Sound clips.
    // NOTE: Audio is not supported yet in the JavaScript version
    // TODO: Should be class wide, not instance wide IMO.
    this.crashSound = null;
    this.explosionSound = null;
    this.fireSound = null;
    this.missileSound = null;
    this.saucerSound = null;
    this.thrustersSound = null;
    this.warpSound = null;

    // Flags for looping sound clips.
    this.thrustersPlaying = false;
    this.saucerPlaying = false;
    this.missilePlaying = false;

    // Values for the offscreen image.
    this.offDimension = null;
    this.offImage = null;
    this.offGraphics = null;

    this.semaphor = 0;

    this.setWindowTitle("Asteroids");
    this.resize(this.screen_width, this.screen_height);

    this.timer = new QTimer(this);
    this.timer.timeout.connect(this, this.run);
    this.timer.start(this.DELAY);

    this.paintEvent = this.repaintScreen;
    // window.onresize  = resize;
    this.keyPressEvent = this.keyDown;
    this.keyReleaseEvent = this.keyUp;

    this.init();

}


/************************************************************************************************
  Main application code -- Methods.
************************************************************************************************/

Asteroids.prototype.getStringWidth = function(s) {
    return s.length * this.FONT_WIDTH;
}

  // Application information.
Asteroids.prototype.getAppInfo = function() {
    return("Asteroids, Original Copyright 1998 by Mike Hall.");
  }

Asteroids.prototype.init = function() {

    /* Graphics g; */
    /* Dimension d; */
    var i;

    // Set clip rectangle
    // setClipRect(0, 0, width, height);

    // Take credit.

    // print(getAppInfo());

    // Find the size of the screen and set the values for sprites.

    /* g = getGraphics(); */
    /* d = size(); */
    /* width = d.width; */
    /* height = d.height; */

    // Generate starry background.

    this.numStars = Math.floor(this.screen_width * this.screen_height / 5000);
    this.stars = []; /* new Point[this.numStars]; */
    for (i = 0; i < this.numStars; i++)
      this.stars[i] = new Point((Math.random() * this.screen_width), (Math.random() * this.screen_height));

    // Create shape for the ship sprite.

    this.ship = new AsteroidsSprite(this);
    this.ship.shape.addPoint(0, -10);
    this.ship.shape.addPoint(7, 10);
    this.ship.shape.addPoint(-7, 10);

    // Create shape for the photon sprites.

    for (i = 0; i < this.MAX_SHOTS; i++) {
      this.photons[i] = new AsteroidsSprite(this);
      this.photons[i].shape.addPoint(1, 1);
      this.photons[i].shape.addPoint(1, -1);
      this.photons[i].shape.addPoint(-1, 1);
      this.photons[i].shape.addPoint(-1, -1);
    }

    // Create shape for the flying saucer.

    this.ufo = new AsteroidsSprite(this);
    this.ufo.shape.addPoint(-15, 0);
    this.ufo.shape.addPoint(-10, -5);
    this.ufo.shape.addPoint(-5, -5);
    this.ufo.shape.addPoint(-5, -9);
    this.ufo.shape.addPoint(5, -9);
    this.ufo.shape.addPoint(5, -5);
    this.ufo.shape.addPoint(10, -5);
    this.ufo.shape.addPoint(15, 0);
    this.ufo.shape.addPoint(10, 5);
    this.ufo.shape.addPoint(-10, 5);

    // Create shape for the guided missile.

    this.missile = new AsteroidsSprite(this);
    this.missile.shape.addPoint(0, -4);
    this.missile.shape.addPoint(1, -3);
    this.missile.shape.addPoint(1, 3);
    this.missile.shape.addPoint(2, 4);
    this.missile.shape.addPoint(-2, 4);
    this.missile.shape.addPoint(-1, 3);
    this.missile.shape.addPoint(-1, -3);

    // Create asteroid sprites.

    for (i = 0; i < this.MAX_ROCKS; i++)
      this.asteroids[i] = new AsteroidsSprite(this);

    // Create explosion sprites.

    for (i = 0; i < this.MAX_SCRAP; i++)
      this.explosions[i] = new AsteroidsSprite(this);

    // Set font data.
    /* NOTE: Fonts are not supported yet
    g.setFont(font);
    fm = g.getFontMetrics();
    fontWidth = fm.getMaxAdvance();
    fontHeight = fm.getHeight();
    */

    // Initialize game data and put us in 'game over' mode.

    this.highScore = 0;
    this.sound = true;
    this.detail = true;
    this.initGame();
    this.endGame();
  }

Asteroids.prototype.initGame = function()
{

    // Initialize game data and sprites.

    this.score = 0;
    this.shipsLeft = this.MAX_SHIPS;
    this.asteroidsSpeed = this.MIN_ROCK_SPEED;
    this.newShipScore = this.NEW_SHIP_POINTS;
    this.newUfoScore = this.NEW_UFO_POINTS;
    this.initShip();
    this.initPhotons();
    this.stopUfo();
    this.stopMissile();
    this.initAsteroids();
    this.initExplosions();
    this.playing = true;
    this.paused = false;
}

Asteroids.prototype.endGame = function()
{

    // Stop ship, flying saucer, guided missile and associated sounds.

    this.playing = false;
    this.stopShip();
    this.stopUfo();
    this.stopMissile();
}

Asteroids.prototype.run = function() {

    // var i, j;
    // var startTime;

    // Load sounds.

    if (!this.loaded) {
      this.loadSounds();
      this.loaded = true;
      // loadThread.stop();
    }

    // This is the main loop.

      if (!this.paused) {

        // Move and process all sprites.

        this.updateShip();
        this.updatePhotons();
        this.updateUfo();
        this.updateMissile();
        this.updateAsteroids();
        this.updateExplosions();

        // Check the score and advance high score, add a new ship or start the flying
        // saucer as necessary.

        if (this.score > this.highScore)
          this.highScore = this.score;
        if (this.score > this.newShipScore) {
          this.newShipScore += this.NEW_SHIP_POINTS;
          this.shipsLeft++;
        }
        if (this.playing && this.score > this.newUfoScore && !this.ufo.active) {
          this.newUfoScore += this.NEW_UFO_POINTS;
          this.ufoPassesLeft = this.UFO_PASSES;
          this.initUfo();
        }

        // If all asteroids have been destroyed create a new batch.

        if (this.asteroidsLeft <= 0) {
            if (--this.asteroidsCounter <= 0)
              this.initAsteroids();
        }

        // Update the screen and set the timer for the next loop.
        //this.repaintScreen();
        this.update();
      }

      // setTimeout(run, DELAY);
  }

Asteroids.prototype.loadSounds = function() {

    // Load all sound clips by playing and immediately stopping them.

    /* NOTE: Sounds are not supported yet
    try {
      crashSound     = getAudioClip(new URL(getDocumentBase(), "crash.au"));
      explosionSound = getAudioClip(new URL(getDocumentBase(), "explosion.au"));
      fireSound      = getAudioClip(new URL(getDocumentBase(), "fire.au"));
      missileSound   = getAudioClip(new URL(getDocumentBase(), "missile.au"));
      saucerSound    = getAudioClip(new URL(getDocumentBase(), "saucer.au"));
      thrustersSound = getAudioClip(new URL(getDocumentBase(), "thrusters.au"));
      warpSound      = getAudioClip(new URL(getDocumentBase(), "warp.au"));
    }
    catch (MalformedURLException e) {}

    crashSound.play();     crashSound.stop();
    explosionSound.play(); explosionSound.stop();
    fireSound.play();      fireSound.stop();
    missileSound.play();   missileSound.stop();
    saucerSound.play();    saucerSound.stop();
    thrustersSound.play(); thrustersSound.stop();
    warpSound.play();      warpSound.stop();
    */
  }

Asteroids.prototype.initShip = function() {

    this.ship.active = true;
    this.ship.angle = 0.0;
    this.ship.deltaAngle = 0.0;
    this.ship.currentX = 0.0;
    this.ship.currentY = 0.0;
    this.ship.deltaX = 0.0;
    this.ship.deltaY = 0.0;
    this.ship.render();
//    if (this.loaded)
//      thrustersSound.stop();
    this.thrustersPlaying = false;

    this.hyperCounter = 0;
  }

Asteroids.prototype.updateShip = function() {

    var dx, dy, limit;

    if (!this.playing)
      return;

    // Rotate the ship if left or right cursor key is down.

    if (this.left) {
      this.ship.angle += Math.PI / 16.0;
      if (this.ship.angle > 2 * Math.PI)
        this.ship.angle -= 2 * Math.PI;
    }
    if (this.right) {
      this.ship.angle -= Math.PI / 16.0;
      if (this.ship.angle < 0)
        this.ship.angle += 2 * Math.PI;
    }

    // Fire thrusters if up or down cursor key is down. Don't let ship go past
    // the speed limit.

    dx = -Math.sin(this.ship.angle);
    dy =  Math.cos(this.ship.angle);
    limit = 0.8 * this.MIN_ROCK_SIZE;
    if (this.up) {
      if (this.ship.deltaX + dx > -limit && this.ship.deltaX + dx < limit)
        this.ship.deltaX += dx;
      if (this.ship.deltaY + dy > -limit && this.ship.deltaY + dy < limit)
        this.ship.deltaY += dy;
    }
    if (this.down) {
      if (this.ship.deltaX - dx > -limit && this.ship.deltaX - dx < limit)
        this.ship.deltaX -= dx;
      if (this.ship.deltaY - dy > -limit && this.ship.deltaY - dy < limit)
        this.ship.deltaY -= dy;
    }

    // Move the ship. If it is currently in hyperspace, advance the countdown.

    if (this.ship.active) {
      this.ship.advance();
      this.ship.render();
      if (this.hyperCounter > 0)
        this.hyperCounter--;
    }

    // Ship is exploding, advance the countdown or create a new ship if it is
    // done exploding. The new ship is added as though it were in hyperspace.
    // (This gives the player time to move the ship if it is in imminent danger.)
    // If that was the last ship, end the game.

    else {
      if (--this.shipCounter <= 0) {
        if (this.shipsLeft > 0) {
          this.initShip();
          this.hyperCounter = this.HYPER_COUNT;
        }
        else
          this.endGame();
      }
    }
  }

Asteroids.prototype.stopShip = function() {

    this.ship.active = false;
    this.shipCounter = this.SCRAP_COUNT;
    if (this.shipsLeft > 0)
      this.shipsLeft--;
//    if (this.loaded)
//      thrustersSound.stop();
    this.thrustersPlaying = false;
  }

Asteroids.prototype.initPhotons = function() {

    for (var i = 0; i < this.MAX_SHOTS; i++) {
      this.photons[i].active = false;
      this.photonCounter[i] = 0;
    }
    this.photonIndex = 0;
  }

Asteroids.prototype.updatePhotons = function() {

    // Move any active photons. Stop it when its counter has expired.

    for (var i = 0; i < this.MAX_SHOTS; i++)
      if (this.photons[i].active) {
        this.photons[i].advance();
        this.photons[i].render();
        if (--this.photonCounter[i] < 0)
          this.photons[i].active = false;
      }
  }

Asteroids.prototype.initUfo = function() {

    var temp;

    // Randomly set flying saucer at left or right edge of the screen.

    this.ufo.active = true;
    this.ufo.currentX = -this.screen_width / 2;
    this.ufo.currentY = Math.random() * this.screen_height;
    this.ufo.deltaX = this.MIN_ROCK_SPEED + Math.random() * (this.MAX_ROCK_SPEED - this.MIN_ROCK_SPEED);
    if (Math.random() < 0.5) {
      this.ufo.deltaX = -this.ufo.deltaX;
      this.ufo.currentX = this.screen_width / 2;
    }
    this.ufo.deltaY = this.MIN_ROCK_SPEED + Math.random() * (this.MAX_ROCK_SPEED - this.MIN_ROCK_SPEED);
    if (Math.random() < 0.5)
      this.ufo.deltaY = -this.ufo.deltaY;
    this.ufo.render();
    this.saucerPlaying = true;
//    if (sound)
//      saucerSound.loop();

    // Set counter for this pass.

    this.ufoCounter = Math.floor(this.screen_width / Math.abs(this.ufo.deltaX));
  }

Asteroids.prototype.updateUfo = function() {

    var i, d;

    // Move the flying saucer and check for collision with a photon. Stop it when its
    // counter has expired.

    if (this.ufo.active) {
      this.ufo.advance();
      this.ufo.render();
      if (--this.ufoCounter <= 0) {
        if (--this.ufoPassesLeft > 0)
          this.initUfo();
        else
          this.stopUfo();
      } else {
        for (i = 0; i < this.MAX_SHOTS; i++)
          if (this.photons[i].active && this.ufo.isColliding(this.photons[i])) {
//            if (sound)
//              crashSound.play();
            this.explode(this.ufo);
            this.stopUfo();
            this.score += this.UFO_POINTS;
          }

          // On occassion, fire a missile at the ship if the saucer is not
          // too close to it.

          d = Math.max(Math.abs(this.ufo.currentX - this.ship.currentX), Math.abs(this.ufo.currentY - this.ship.currentY));
          if (this.ship.active && this.hyperCounter <= 0 && this.ufo.active && !this.missile.active &&
              d > 4 * this.MAX_ROCK_SIZE && Math.random() < .03)
            this.initMissile();
       }
    }
  }

Asteroids.prototype.stopUfo = function() {

    this.ufo.active = false;
    this.ufoCounter = 0;
    this.ufoPassesLeft = 0;
//    if (this.loaded)
//      saucerSound.stop();
    this.saucerPlaying = false;
  }

Asteroids.prototype.initMissile = function() {

    this.missile.active = true;
    this.missile.angle = 0.0;
    this.missile.deltaAngle = 0.0;
    this.missile.currentX = this.ufo.currentX;
    this.missile.currentY = this.ufo.currentY;
    this.missile.deltaX = 0.0;
    this.missile.deltaY = 0.0;
    this.missile.render();
    this.missileCounter = 3 * Math.max(this.screen_width, this.screen_height) / this.MIN_ROCK_SIZE;
//    if (sound)
//      missileSound.loop();
    this.missilePlaying = true;
  }

Asteroids.prototype.updateMissile = function() {

    var i;

    // Move the guided missile and check for collision with ship or photon. Stop it when its
    // counter has expired.

    if (this.missile.active) {
      if (--this.missileCounter <= 0)
        stopMissile();
      else {
        this.guideMissile();
        this.missile.advance();
        this.missile.render();
        for (i = 0; i < this.MAX_SHOTS; i++)
          if (this.photons[i].active && this.missile.isColliding(this.photons[i])) {
//            if (sound)
//              crashSound.play();
            explode(this.missile);
            stopMissile();
            this.score += this.MISSILE_POINTS;
          }
        if (this.missile.active && this.ship.active && this.hyperCounter <= 0 && this.ship.isColliding(this.missile)) {
//          if (sound)
//            crashSound.play();
          this.explode(this.ship);
          this.stopShip();
          this.stopUfo();
          this.stopMissile();
        }
      }
    }
  }

Asteroids.prototype.guideMissile = function() {

    /* double */ var dx, dy, angle;

    if (!this.ship.active || this.hyperCounter > 0)
      return;

    // Find the angle needed to hit the ship.

    dx = this.ship.currentX - this.missile.currentX;
    dy = this.ship.currentY - this.missile.currentY;
    if (dx == 0 && dy == 0)
      angle = 0;
    if (dx == 0) {
      if (dy < 0)
        angle = -Math.PI / 2;
      else
        angle = Math.PI / 2;
    }
    else {
      angle = Math.atan(Math.abs(dy / dx));
      if (dy > 0)
        angle = -angle;
      if (dx < 0)
        angle = Math.PI - angle;
    }

    // Adjust angle for screen coordinates.

    this.missile.angle = angle - Math.PI / 2;

    // Change the this.missile's angle so that it points toward the ship.

    this.missile.deltaX = this.MIN_ROCK_SIZE / 3 * -Math.sin(this.missile.angle);
    this.missile.deltaY = this.MIN_ROCK_SIZE / 3 *  Math.cos(this.missile.angle);
  }

Asteroids.prototype.stopMissile = function() {

    this.missile.active = false;
    this.missileCounter = 0;
//    if (this.loaded)
//      missileSound.stop();
    this.missilePlaying = false;
  }

Asteroids.prototype.initAsteroids = function() {

    var i, j;
    var s;
    /* double */ var theta, r;
    var x, y;

    // Create random shapes, positions and movements for each asteroid.

    for (i = 0; i < this.MAX_ROCKS; i++) {

      // Create a jagged shape for the asteroid and give it a random rotation.

      this.asteroids[i].shape = new Polygon();
      s = this.MIN_ROCK_SIDES + (Math.random() * (this.MAX_ROCK_SIDES - this.MIN_ROCK_SIDES));
      for (j = 0; j < s; j ++) {
        theta = 2 * Math.PI / s * j;
        r = this.MIN_ROCK_SIZE + (Math.random() * (this.MAX_ROCK_SIZE - this.MIN_ROCK_SIZE));
        x = -Math.round(r * Math.sin(theta));
        y =  Math.round(r * Math.cos(theta));
        this.asteroids[i].shape.addPoint(x, y);
      }
      this.asteroids[i].active = true;
      this.asteroids[i].angle = 0.0;
      this.asteroids[i].deltaAngle = (Math.random() - 0.5) / 10;

      // Place the asteroid at one edge of the screen.

      if (Math.random() < 0.5) {
        this.asteroids[i].currentX = -this.screen_width / 2;
        if (Math.random() < 0.5)
          this.asteroids[i].currentX = this.screen_width / 2;
        this.asteroids[i].currentY = Math.random() * this.screen_height;
      }
      else {
        this.asteroids[i].currentX = Math.random() * this.screen_width;
        this.asteroids[i].currentY = -this.screen_height / 2;
        if (Math.random() < 0.5)
          this.asteroids[i].currentY = this.screen_height / 2;
      }

      // Set a random motion for the asteroid.

      this.asteroids[i].deltaX = Math.random() * this.asteroidsSpeed;
      if (Math.random() < 0.5)
        this.asteroids[i].deltaX = -this.asteroids[i].deltaX;
      this.asteroids[i].deltaY = Math.random() * this.asteroidsSpeed;
      if (Math.random() < 0.5)
        this.asteroids[i].deltaY = -this.asteroids[i].deltaY;

      this.asteroids[i].render();
      this.asteroidIsSmall[i] = false;
    }

    this.asteroidsCounter = this.STORM_PAUSE;
    this.asteroidsLeft = this.MAX_ROCKS;
    if (this.asteroidsSpeed < this.MAX_ROCK_SPEED)
      this.asteroidsSpeed++;
  }

Asteroids.prototype.initSmallAsteroids = function(n) {

    var count;
    var i, j;
    var s;
    /* double */ var tempX, tempY;
    /* double */ var theta, r;
    var x, y;

    // Create one or two smaller asteroids from a larger one using inactive asteroids. The new
    // asteroids will be placed in the same position as the old one but will have a new, smaller
    // shape and new, randomly generated movements.

    count = 0;
    i = 0;
    tempX = this.asteroids[n].currentX;
    tempY = this.asteroids[n].currentY;
    do {
      if (!this.asteroids[i].active) {
        this.asteroids[i].shape = new Polygon();
        s = this.MIN_ROCK_SIDES + (Math.random() * (this.MAX_ROCK_SIDES - this.MIN_ROCK_SIDES));
        for (j = 0; j < s; j ++) {
          theta = 2 * Math.PI / s * j;
          r = (this.MIN_ROCK_SIZE + (Math.random() * (this.MAX_ROCK_SIZE - this.MIN_ROCK_SIZE))) / 2;
          x = -Math.round(r * Math.sin(theta));
          y =  Math.round(r * Math.cos(theta));
          this.asteroids[i].shape.addPoint(x, y);
        }
        this.asteroids[i].active = true;
        this.asteroids[i].angle = 0.0;
        this.asteroids[i].deltaAngle = (Math.random() - 0.5) / 10;
        this.asteroids[i].currentX = tempX;
        this.asteroids[i].currentY = tempY;
        this.asteroids[i].deltaX = Math.random() * 2 * this.asteroidsSpeed - this.asteroidsSpeed;
        this.asteroids[i].deltaY = Math.random() * 2 * this.asteroidsSpeed - this.asteroidsSpeed;
        this.asteroids[i].render();
        this.asteroidIsSmall[i] = true;
        count++;
        this.asteroidsLeft++;
      }
      i++;
    } while (i < this.MAX_ROCKS && count < 2);
  }

Asteroids.prototype.updateAsteroids = function() {

    var i, j;

    // Move any active asteroids and check for collisions.

    for (i = 0; i < this.MAX_ROCKS; i++)
      if (this.asteroids[i].active) {
        this.asteroids[i].advance();
        this.asteroids[i].render();

        // If hit by photon, kill asteroid and advance score. If asteroid is large,
        // make some smaller ones to replace it.

        for (j = 0; j < this.MAX_SHOTS; j++)
          if (this.photons[j].active && this.asteroids[i].active && this.asteroids[i].isColliding(this.photons[j])) {
            this.asteroidsLeft--;
            this.asteroids[i].active = false;
            this.photons[j].active = false;
//            if (sound)
//              explosionSound.play();
            this.explode(this.asteroids[i]);
            if (!this.asteroidIsSmall[i]) {
              this.score += this.BIG_POINTS;
              this.initSmallAsteroids(i);
            }
            else
              this.score += this.SMALL_POINTS;
          }

        // If the ship is not in hyperspace, see if it is hit.

        if (this.ship.active && this.hyperCounter <= 0 && this.asteroids[i].active && this.asteroids[i].isColliding(this.ship)) {
//          if (sound)
//            crashSound.play();
          this.explode(this.ship);
          this.stopShip();
          this.stopUfo();
          this.stopMissile();
        }
    }
  }

Asteroids.prototype.initExplosions = function() {

    for (var i = 0; i < this.MAX_SCRAP; i++) {
      this.explosions[i].shape = new Polygon();
      this.explosions[i].active = false;
      this.explosionCounter[i] = 0;
    }
    this.explosionIndex = 0;
  }

Asteroids.prototype.explode = function(/* AsteroidsSprite */ s) {

    var c, i, j;

    // Create sprites for explosion animation. The each individual line segment of the given sprite
    // is used to create a new sprite that will move outward  from the sprite's original position
    // with a random rotation.

    s.render();
    c = 2;
    if (this.detail || s.sprite.npoints < 6)
      c = 1;
    for (i = 0; i < s.sprite.npoints; i += c) {
      this.explosionIndex++;
      if (this.explosionIndex >= this.MAX_SCRAP)
        this.explosionIndex = 0;
      this.explosions[this.explosionIndex].active = true;
      this.explosions[this.explosionIndex].shape = new Polygon();
      this.explosions[this.explosionIndex].shape.addPoint(s.shape.xpoints[i], s.shape.ypoints[i]);
      j = i + 1;
      if (j >= s.sprite.npoints)
        j -= s.sprite.npoints;
      this.explosions[this.explosionIndex].shape.addPoint(s.shape.xpoints[j], s.shape.ypoints[j]);
      this.explosions[this.explosionIndex].angle = s.angle;
      this.explosions[this.explosionIndex].deltaAngle = (Math.random() * 2 * Math.PI - Math.PI) / 15;
      this.explosions[this.explosionIndex].currentX = s.currentX;
      this.explosions[this.explosionIndex].currentY = s.currentY;
      this.explosions[this.explosionIndex].deltaX = -s.shape.xpoints[i] / 5;
      this.explosions[this.explosionIndex].deltaY = -s.shape.ypoints[i] / 5;
      this.explosionCounter[this.explosionIndex] = this.SCRAP_COUNT;
    }
  }

Asteroids.prototype.updateExplosions = function() {

    // Move any active explosion debris. Stop explosion when its counter has expired.

    for (var i = 0; i < this.MAX_SCRAP; i++)
      if (this.explosions[i].active) {
        this.explosions[i].advance();
        this.explosions[i].render();
        if (--this.explosionCounter[i] < 0)
          this.explosions[i].active = false;
      }
  }

Asteroids.prototype.repaintScreen = function() {

    if (this.semaphor > 0) {
      print("Semaphor blocked repainting!");
      return;
    }
    this.semaphor = 1;

    // Dimension d = size();
    var i;
    var c;
    /* String */ var s;

    // Create the offscreen graphics context, if no good one exists.
    // NOTE: Qt handles double-buffering automatically, 
    // so this code is not needed
    /*
    if (offGraphics == null || d.width != offDimension.width || d.height != offDimension.height) {
      offDimension = d;
      offImage = createImage(d.width, d.height);
      offGraphics = offImage.getGraphics();
    }
    */

    // Fill in background and stars.

    var painter = new QPainter();

    painter.begin(this);
    painter.setRenderHint(QPainter.Antialiasing);
    // painter.setPen(new QPen());
    painter.setClipRect(0, 0, this.screen_width, this.screen_height);

    var blackBrush = new QBrush(this.BLACK_COLOR);
    painter.setBrush(blackBrush);

    painter.fillRect(0, 0, this.screen_width, this.screen_height, blackBrush);
    if (this.detail) {
      painter.setPen(this.YELLOW_COLOR);
      for (i = 0; i < this.numStars; i++)
        painter.drawLine(this.stars[i].x, this.stars[i].y, this.stars[i].x, this.stars[i].y+1);
    }

    // Draw photon bullets.

    /* offGraphics. */ painter.setPen(this.YELLOW_COLOR);
    for (i = 0; i < this.MAX_SHOTS; i++)
      if (this.photons[i].active)
        this.photons[i].sprite.draw(painter);

    // Draw the guided missile, counter is used to quickly fade color to black when near expiration.

    c = Math.min(this.missileCounter * 24, 255);
    painter.setPen(new QColor(c, c, c));
    if (this.missile.active) {
      this.missile.sprite.draw(painter);
      painter.drawLine(this.missile.sprite.xpoints[this.missile.sprite.npoints - 1], this.missile.sprite.ypoints[this.missile.sprite.npoints - 1],
                       this.missile.sprite.xpoints[0], this.missile.sprite.ypoints[0]);
    }

    // Draw the asteroids.

    for (i = 0; i < this.MAX_ROCKS; i++) {
      if (this.asteroids[i].active) {
        if (this.detail) {
          // setColor(Color.black);
          // this.asteroids[i].sprite.fill(painter); Not needed in Qt
        }
        painter.setPen(this.YELLOW_COLOR);
        this.asteroids[i].sprite.draw(painter);
        painter.drawLine(this.asteroids[i].sprite.xpoints[this.asteroids[i].sprite.npoints - 1], this.asteroids[i].sprite.ypoints[this.asteroids[i].sprite.npoints - 1],
                         this.asteroids[i].sprite.xpoints[0], this.asteroids[i].sprite.ypoints[0]);
      }
    }

    // Draw the flying saucer.

    if (this.ufo.active) {
      if (this.detail) {
        // setColor(Color.black);
        // this.ufo.sprite.fill(painter); Not needed in Qt
      }
      painter.setPen(this.YELLOW_COLOR);
      this.ufo.sprite.draw(painter);
      painter.drawLine(this.ufo.sprite.xpoints[this.ufo.sprite.npoints - 1], this.ufo.sprite.ypoints[this.ufo.sprite.npoints - 1],
                       this.ufo.sprite.xpoints[0], this.ufo.sprite.ypoints[0]);
    }

    // Draw the ship, counter is used to fade color to white on hyperspace.

    c = 255 - (255 / this.HYPER_COUNT) * this.hyperCounter;
    if (this.ship.active) {
      if (this.detail && this.hyperCounter == 0) {
        // setColor(Color.black);
        // ship.sprite.fill(painter); Not needed in Qt
      }
      painter.setPen(new QColor(c, c, c));
      this.ship.sprite.draw(painter);
      painter.drawLine(this.ship.sprite.xpoints[this.ship.sprite.npoints - 1], this.ship.sprite.ypoints[this.ship.sprite.npoints - 1],
                       this.ship.sprite.xpoints[0], this.ship.sprite.ypoints[0]);
    }

    // Draw any explosion debris, counters are used to fade color to black.

    for (i = 0; i < this.MAX_SCRAP; i++)
      if (this.explosions[i].active) {
        c = (255 / this.SCRAP_COUNT) * this.explosionCounter [i];
        painter.setPen(new QColor(c, c, c));
        this.explosions[i].sprite.draw(painter);
      }

    // Display status and messages.

    // setFont(font);
    painter.setPen(this.YELLOW_COLOR);

    painter.drawText(this.FONT_WIDTH, 20, "Score: " + this.score);
    painter.drawText(this.FONT_WIDTH, this.screen_height - this.FONT_HEIGHT, "Ships: " + this.shipsLeft);
    s = "High: " + this.highScore;
    painter.drawText(this.screen_width - (this.FONT_WIDTH + this.getStringWidth(s)), 20, s);
    if (!this.sound) {
      s = "Mute";
      painter.drawText(this.screen_width - (this.FONT_WIDTH + this.getStringWidth(s)), this.screen_height - this.FONT_HEIGHT, s);
    }

    if (!this.playing) {
      s = "A S T E R O I D S";
      painter.drawText((this.screen_width - this.getStringWidth(s)) / 2, this.screen_height / 2, s);
      s = "Copyright 1998 by Mike Hall";
      painter.drawText((this.screen_width - this.getStringWidth(s)) / 2, this.screen_height / 2 + this.FONT_HEIGHT, s);
        s = "'S' = Start.  'P' = Pause.  'H' = Hyperspace.";
        painter.drawText((this.screen_width - this.getStringWidth(s)) / 2, this.screen_height / 2 + this.FONT_HEIGHT*3, s);
      //if (!this.loaded) {
      //  s = "Loading sounds...";
      //  /* offGraphics. */ painter.drawText(s, (width - this.getStringWidth(s)) / 2, this.screen_height / 4);
      //}
      //else {
        s = "Game Over";
        painter.drawText((this.screen_width - this.getStringWidth(s)) / 2, this.screen_height / 4, s);
      //}
    }
    else if (this.paused) {
      s = "Game Paused";
      painter.setPen(this.YELLOW_COLOR);
      painter.drawText((this.screen_width - this.getStringWidth(s)) / 2, this.screen_height / 2, s);
    }

    // Copy the off screen buffer to the screen.
    // g.drawImage(offImage, 0, 0, this);

    painter.end();
    this.semaphor = 0;
  }

/************************************************************************************************
  Main application code -- Event Handlers
************************************************************************************************/

  /* Keycodes */
Asteroids.prototype.KEY_LEFT  = 16777234; // 37;
Asteroids.prototype.KEY_RIGHT = 16777236; // 39;
Asteroids.prototype.KEY_UP    = 16777235; // 38;
Asteroids.prototype.KEY_DOWN  = 16777237; // 40;
Asteroids.prototype.KEY_SPACE = 32;
Asteroids.prototype.KEY_S     = 83; // Start
Asteroids.prototype.KEY_M     = 77; // Mute
Asteroids.prototype.KEY_D     = 68; // Detail
Asteroids.prototype.KEY_P     = 80; // Pause
Asteroids.prototype.KEY_H     = 72; // Hyperspace

Asteroids.prototype.keyDown = function(event) {

    var key = event.key();

    // Check if any cursor keys have been pressed and set flags.

    if (key == this.KEY_LEFT)
      this.left = true;
    if (key == this.KEY_RIGHT)
      this.right = true;
    if (key == this.KEY_UP)
      this.up = true;
    if (key == this.KEY_DOWN)
      this.down = true;

    if ((this.up || this.down) && this.ship.active && !this.thrustersPlaying) {
//      if (sound && !this.paused)
//        thrustersSound.loop();
      this.thrustersPlaying = true;
    }

    // Spacebar: fire a photon and start its counter.

    if (key == this.KEY_SPACE && this.ship.active) {
//      if (sound & !this.paused)
//        fireSound.play();
      this.photonIndex++;
      if (this.photonIndex >= this.MAX_SHOTS)
        this.photonIndex = 0;
      this.photons[this.photonIndex].active = true;
      this.photons[this.photonIndex].currentX = this.ship.currentX;
      this.photons[this.photonIndex].currentY = this.ship.currentY;
      this.photons[this.photonIndex].deltaX = this.MIN_ROCK_SIZE * -Math.sin(this.ship.angle);
      this.photons[this.photonIndex].deltaY = this.MIN_ROCK_SIZE *  Math.cos(this.ship.angle);
      this.photonCounter[this.photonIndex] = Math.min(this.screen_width, this.screen_height) / this.MIN_ROCK_SIZE;
    }

    // 'H' key: warp ship into hyperspace by moving to a random location and starting counter.

    if (key == this.KEY_H && this.ship.active && this.hyperCounter <= 0) {
      this.ship.currentX = Math.random() * this.screen_width;
      this.ship.currentX = Math.random() * this.screen_height;
      this.hyperCounter = this.HYPER_COUNT;
//      if (sound & !this.paused)
//        warpSound.play();
    }

    // 'P' key: toggle pause mode and start or stop any active looping sound clips.

    if (key == this.KEY_P) {
      if (this.paused) {
//        if (sound && this.missilePlaying)
//          this.missileSound.loop();
//        if (sound && saucerPlaying)
//          saucerSound.loop();
//        if (sound && thrustersPlaying)
//          thrustersSound.loop();
      }
      else {
//        if (this.missilePlaying)
//          this.missileSound.stop();
//        if (saucerPlaying)
//          saucerSound.stop();
//        if (thrustersPlaying)
//          thrustersSound.stop();
      }
      this.paused = !this.paused;
    }

    // 'M' key: toggle sound on or off and stop any looping sound clips.

    if (key == this.KEY_M && this.loaded) {
/*
      if (sound) {
        crashSound.stop();
        explosionSound.stop();
        fireSound.stop();
        this.missileSound.stop();
        saucerSound.stop();
        thrustersSound.stop();
        warpSound.stop();
      }
      else {
        if (this.missilePlaying && !this.paused)
          this.missileSound.loop();
        if (saucerPlaying && !this.paused)
          saucerSound.loop();
        if (thrustersPlaying && !this.paused)
          thrustersSound.loop();
      }
      sound = !sound;
*/
    }

    // 'D' key: toggle graphics detail on or off.

    if (key == this.KEY_D) {
      this.detail = !this.detail;
    }

    // 'S' key: start the game, if not already in progress.

    if (key == this.KEY_S && this.loaded && !this.playing)
      this.initGame();

    return true;
  }

Asteroids.prototype.keyUp = function(event) {

    var key = event.key();

    // Check if any cursor keys where released and set flags.

    if (key == this.KEY_LEFT)
      this.left = false;
    if (key == this.KEY_RIGHT)
      this.right = false;
    if (key == this.KEY_UP)
      this.up = false;
    if (key == this.KEY_DOWN)
      this.down = false;

    if (!this.up && !this.down && this.thrustersPlaying) {
//      thrustersSound.stop();
      this.thrustersPlaying = false;
    }

    return true;
  }

/*
  function resize(event) {
    clearScreen();

    // Set clip rectangle
    setClipRect(0, 0, width, height);

    asteroidGame.repaint();
  }
*/

/************************************************************************************************
  Main application code -- Startup!
************************************************************************************************/

/* To use the application in standalone mode, remove the comment around this code
var asteroidGame = new Asteroids();

asteroidGame.show();
QCoreApplication.exec();
*/

