/*========================================================================
 * Copyright (c) 2008-2009 Sun Microsystems Laboratories
 * Copyright (c) 2008-2009 Tampere University of Technology
 * 
 * Permission is hereby granted, free of charge, to any person 
 * obtaining a copy of this software and associated documentation 
 * files (the "Software"), to deal in the Software without restriction, 
 * including without limitation the rights to use, copy, modify, merge, 
 * publish, distribute, sublicense, and/or sell copies of the Software, 
 * and to permit persons to whom the Software is furnished to do so, 
 * subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.  
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY 
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE 
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *======================================================================*/

/*======================================================================
 * QtInspector.js: Simple visual JavaScript object inspector / browser
 *====================================================================*/

// TODO: Inspector only works correctly when starting the inspection
// from the Global object; generalize the code later!

if (!Global) {
    var Global = this;
}

/*======================================================================
 * QtInspector class
 *====================================================================*/

// Define the QtInspector class as a subclass of QWidget
function QtInspector(parent, objectToView, objectTitle) {
    QWidget.call(this, parent);

    // Default inspector size
    this.defaultWidth = 600;
    this.defaultHeight = 300;

    this.objectToView = (objectToView !== undefined) ? objectToView : Global;
    this.objectTitle  = (objectTitle  !== undefined) ? objectTitle  : "Global";

    // Set window (QWidget) attributes
    this.setWindowTitle("QtInspector");
    this.resize(this.defaultWidth, this.defaultHeight);
    this.viewMode = QtInspector.INSPECT_VARIABLES; // INSPECT_ALL

    this.modeSelector = new QComboBox(this);
    this.modeSelector.addItem(QtInspector.INSPECT_ALL, null);
    this.modeSelector.addItem(QtInspector.INSPECT_VARIABLES, null);
    this.modeSelector.addItem(QtInspector.INSPECT_FUNCTIONS, null);
    this.modeSelector.setCurrentIndex(1); // INSPECT_VARIABLES
    //this.modeSelector.show();

    this.refreshButton = new QPushButton("Refresh", this);

    this.sortBox = new QCheckBox("Enable Sorting", this);

    this.clipBox = new QCheckBox("Clip Long Lines", this);
    this.clipBox.setChecked(true);

    // Create hboxLayout and add some widgets into it
    var hboxLayout = new QHBoxLayout();
    hboxLayout.addWidget(this.refreshButton, 0, 0);
    hboxLayout.addSpacing(10);
    hboxLayout.addWidget(this.sortBox, 0, 0);
    hboxLayout.addSpacing(10);
    hboxLayout.addWidget(this.clipBox, 0, 0);
    hboxLayout.addStretch();
    hboxLayout.addWidget(this.modeSelector, 0, 0);
    
    // Create vboxLayout and add hboxLayout and QTreeWidget into it
    this.layout = new QVBoxLayout(this);
    this.layout.addLayout(hboxLayout);
    this.layout.setContentsMargins(2, 2, 2, 2);
    // QTreeWidget is added later

    // Generate data to display
    this.rebuildEntireTree();
    // this.show();
    return this;
}

// QtInspector is a subclass of QWidget
QtInspector.prototype = new QWidget();

/*======================================================================
 * Constants
 *====================================================================*/

// View modes
QtInspector.INSPECT_ALL       = "Show All";
QtInspector.INSPECT_VARIABLES = "Show Variables Only";
QtInspector.INSPECT_FUNCTIONS = "Show Functions Only";

/*======================================================================
 * Simple reflection interface
 *====================================================================*/

// Find all the variables of an object
QtInspector.prototype.dumpVariables = function(obj, parentTreeItem, fullPath) {

    // Count the number of properties found
    var counter = 0;

    for (var name in obj) {
        var object = obj[name];

        if (!(object instanceof Function) && (object !== Global)) {
            counter++;

            var treeItem = new QTreeWidgetItem(parentTreeItem);
            parentTreeItem.addChild(treeItem);
            treeItem.setChildIndicatorPolicy(QTreeWidgetItem.ShowIndicator);

            // Set data in column 0 (object's name)
            treeItem.setText(0, name);

            // VERY Qt-specific: Unfortunately, we cannot easily attach
            // a JavaScript pointer to a QTreeWidgetItem so that we could
            // point back to the original object from the widget easily.
            // We need to store the full path to the object as a string
            // and then eval it back later :-(
            var newPath = fullPath;
            if (typeof name == "number") newPath += "[" + name + "]";
            else newPath += "." + name; 

            // We use Qt's 'setData' function to store the path string.
            // NOTE: The path is not visible in the tree
            treeItem.setData(2, Qt.UserRole, newPath);

            // Eliminate overly large arrays from the tree
            if (typeof object == "array" && object.length > 256) {
                 treeItem.setText(1, "[... very long Array (" + object.length + " items) ...]");
                 continue;
            }

            // Eliminate overly large strings from the tree
            if (typeof object == "string" && object.length > 256) {
                 treeItem.setText(1, object.substring(0, 256));
                 continue;
            }

            // A hack to get rid of a Qt-specific error related to QLocale objects
            if (object instanceof QLocale) { 
                 treeItem.setText(1, "(a QLocale object)");
                 continue;
            } 

            treeItem.setText(1, this.clipValue(object)); // Set data in column 1 (object's value)

            // If the object does not have any expandable properties, 
            // remove the expand button from the item
            if (!object) {
                treeItem.setChildIndicatorPolicy(QTreeWidgetItem.DontShowIndicator);
            } else if (object === true || object === false) {
                treeItem.setChildIndicatorPolicy(QTreeWidgetItem.DontShowIndicator);
            } else if (typeof object == "number") {
                treeItem.setChildIndicatorPolicy(QTreeWidgetItem.DontShowIndicator);
            } else if (typeof obj == "string" && typeof object == "string") {
                treeItem.setChildIndicatorPolicy(QTreeWidgetItem.DontShowIndicatorWhenChildless);
            }
        }
    }
    
    // Return the number of properties found
    return counter;
}

// Find all the functions of an object
QtInspector.prototype.dumpFunctions = function(obj, parentTreeItem, fullPath) {
    var counter = 0;

    for (var name in obj) {
        if (obj[name] instanceof Function) {
            counter++;
            var treeItem = new QTreeWidgetItem(parentTreeItem);
            parentTreeItem.addChild(treeItem);

            // Column 0: name.  Column 1: value.  Column 2: hidden data (a string)
            treeItem.setText(0, name);
            treeItem.setText(1, this.clipValue(obj[name]));
            treeItem.setData(2, Qt.UserRole, fullPath + "." + name);
        }
    }
    return counter;
}

// Find all the variables and functions of an object
QtInspector.prototype.dumpAll = function(obj, parentTreeItem, fullPath) {
    var counter = 0;
    counter += this.dumpVariables(obj, parentTreeItem, fullPath);
    counter += this.dumpFunctions(obj, parentTreeItem, fullPath);
    return counter;
}

// If the clipBox CheckBox is set, cut an overly 
// large string and eliminate all newlines and CRs
QtInspector.prototype.clipValue = function(object) {
    if (!object) return object;

    var string = object.toString();
    if (!this.clipBox.checked) return string;
    
    // Eliminate CRs and newlines
    // RegExp doesn't seem to work
    // var newstr = string.substring(0, 80).replace(/[\r\n\v]*/, "");

    var substr = string.substring(0, 80);
    var newstr = "";
    for (var i = 0; i < substr.length; i++) {
        var ch = substr.charAt(i); 
        if (ch == "\n") newstr += " ";
        else newstr += ch;  
    }

    if (string.length > 80)
        return newstr + " ...";
    else return newstr;
}

// Build the tree for the given object
QtInspector.prototype.buildTree = function(objectToView, parentTreeItem, fullPath) {

    // Remove all the children from the tree to avoid duplicates
    parentTreeItem.takeChildren();
    var counter = 0;

    switch (this.viewMode) {

    case QtInspector.INSPECT_ALL:
        counter = this.dumpAll(objectToView, parentTreeItem, fullPath);
        break;

    case QtInspector.INSPECT_VARIABLES:
        counter = this.dumpVariables(objectToView, parentTreeItem, fullPath);
        break;

    case QtInspector.INSPECT_FUNCTIONS:
        counter = this.dumpFunctions(objectToView, parentTreeItem, fullPath);
        break;

    }

    // If the object has no displayable properties, remove the expand button
    // to prevent the user from unnecessarily clicking empty items
    if (counter == 0) {
        parentTreeItem.setChildIndicatorPolicy(QTreeWidgetItem.DontShowIndicator);
    } else {
        // Optional: Display variables in sorted order.
        // Skip arrays and strings, because their items, 
        // if sorted, would be displayed in meaningless order
        if (this.sortBox.checked && 
            (typeof objectToView != "array") && (typeof objectToView != "string")) {
            parentTreeItem.sortChildren(0, Qt.AscendingOrder);
        }
    }
    
    return counter;
}

// Event handler for expanding a single object in the tree
QtInspector.prototype.expandObject = function(treeItem) {

    // VERY Qt-specific: Unfortunately, we cannot easily attach
    // a JavaScript pointer to a QTreeWidgetItem so that we could
    // get a pointer back to the original object from the widget easily.
    // To avoid this problem, we have stored the full path to the object
    // to a hidden data field.  We can then 'eval' the path string to 
    // obtain a pointer to the original object.
    var targetObjectPath = treeItem.data(2, Qt.UserRole);
    var targetObject = this.pathToObject(targetObjectPath);

    if (targetObject) this.buildTree(targetObject, treeItem, targetObjectPath);
    else print("targetObject was null"); // Sanity check
}

// Given a path name (such as "x.y.z") as a string, 
// get a pointer to the corresponding object
QtInspector.prototype.pathToObject = function(path) {
    // This operation is plain stupid!
    // However, Qt does not allow us to carry pointers
    // in QTreeWidgetItems without losing their type info.
    // So, we get the object back by evaluating it from a string.
    var object; 
    
    try {
        object = eval(path);
    } catch(e) {
        // Squelch all error messages; this should not fail anyway
        // Null is returned if it does fail
    }

    return object;
}

// Rebuild the entire tree displayed in the browser
QtInspector.prototype.rebuildEntireTree = function() {

    // Get rid of the previous tree, if any
    if (this.tree) {
        this.tree.clear();
    } else {

        // Create the tree widget
        var tree = this.tree = new QTreeWidget(this);
        tree.columnCount = 2;
        tree.setColumnWidth(0, this.width/3);
        tree.setHeaderLabels(["Name", "Value"]);

        // Add tree to this.layout
        this.layout.addWidget(tree, Qt.AlignCenter, 0);
    }

    // Sorting does not work well; avoid it for now
    if (this.sortBox.checked) 
         this.tree.sortingEnabled = true;
    else this.tree.sortingEnabled = false;

    // Create the initial branch of the tree for the Global namespace
    this.globalTreeItem = new QTreeWidgetItem(this.tree);
    this.globalTreeItem.setText(0, this.objectTitle /* "Global" */);

    // Initialize the inspector state and auto-expand the first level
    var counter = this.buildTree(this.objectToView, this.globalTreeItem, this.objectTitle /* "Global" */);
    this.globalTreeItem.setExpanded(true);

    // Read the Qt-specific comments about 'setData' earlier
    this.globalTreeItem.setText(1, "(" + counter + " properties displayed)");
    this.globalTreeItem.setData(2, Qt.UserRole, "Global");

    // Initialize Qt-specific event handlers
    this.initializeEventHandlers();
}


// Invoke an item from the 'modeSelector' combobox
QtInspector.prototype.selectBrowserMode = function(index) {

    switch (index) {
    
    case QtInspector.INSPECT_ALL: 
        if (this.viewMode !== QtInspector.INSPECT_ALL) {
            this.viewMode = QtInspector.INSPECT_ALL;
            this.rebuildEntireTree();
        }
        break;
                
    case QtInspector.INSPECT_VARIABLES:
        if (this.viewMode !== QtInspector.INSPECT_VARIABLES) {
            this.viewMode = QtInspector.INSPECT_VARIABLES;
            this.rebuildEntireTree();
        }
        break;
        
    case QtInspector.INSPECT_FUNCTIONS:
        if (this.viewMode !== QtInspector.INSPECT_FUNCTIONS) {
            this.viewMode = QtInspector.INSPECT_FUNCTIONS;
            this.rebuildEntireTree();
        }
        break;    
    }

}

/*==============================================================================
 * Event handling support
 *============================================================================*/

// Event handler initialization
QtInspector.prototype.initializeEventHandlers = function() {

    // Analyze a subobject when expanding a tree branch
    this.tree.itemExpanded.connect(this, this.expandObject);

    // Connect the refresh button to its action
    this.refreshButton.pressed.connect(this, this.rebuildEntireTree);

    // Connect the combobox to its actions
    // NOTE: Qt 4.5 QComboBox has two different 'valueChanged' methods!
    this.modeSelector['activated(QString)'].connect(this, this.selectBrowserMode);

    // Connect the check boxes to the rebuild action
    this.sortBox.toggled.connect(this, this.rebuildEntireTree);
    this.clipBox.toggled.connect(this, this.rebuildEntireTree);

    // Enable window resizing
    this.resizeEvent = this.resizeWindow;
}

/*==============================================================================
 * Instantiation
 *============================================================================*/

// var qtInspector = new QtInspector();
// qtInspector.show();

